# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from ovs.stream import Stream
from ovsdbapp.backend import ovs_idl
from ovsdbapp.backend.ovs_idl import command
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import rowview
from ovsdbapp import event
from ovsdbapp.schema.ovn_northbound import impl_idl as nb_impl_idl
from ovsdbapp.schema.ovn_southbound import impl_idl as sb_impl_idl

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent import exceptions
from ovn_bgp_agent.utils import helpers

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class OvnIdl(connection.OvsdbIdl):
    def __init__(self, driver, remote, schema, **kwargs):
        super(OvnIdl, self).__init__(remote, schema, **kwargs)
        self.driver = driver
        self.notify_handler = OvnDbNotifyHandler(driver)

    def notify(self, event, row, updates=None):
        self.notify_handler.notify(event, row, updates)


class OvnDbNotifyHandler(event.RowEventHandler):
    def __init__(self, driver):
        super(OvnDbNotifyHandler, self).__init__()
        self.driver = driver


class OvnNbIdl(OvnIdl):
    SCHEMA = 'OVN_Northbound'

    def __init__(self, connection_string, events=None, tables=None,
                 leader_only=False):
        if connection_string.startswith("ssl"):
            self._check_and_set_ssl_files(self.SCHEMA)
        helper = self._get_ovsdb_helper(connection_string)
        self._events = events
        if tables is None:
            tables = ('Logical_Switch_Port', 'NAT', 'NB_Global')
        for table in tables:
            helper.register_table(table)
        super(OvnNbIdl, self).__init__(
            None, connection_string, helper, leader_only=leader_only)

    def _get_ovsdb_helper(self, connection_string):
        return idlutils.get_schema_helper(connection_string, self.SCHEMA)

    def _check_and_set_ssl_files(self, schema_name):
        priv_key_file = CONF.ovn.ovn_nb_private_key
        cert_file = CONF.ovn.ovn_nb_certificate
        ca_cert_file = CONF.ovn.ovn_nb_ca_cert

        if priv_key_file:
            Stream.ssl_set_private_key_file(priv_key_file)

        if cert_file:
            Stream.ssl_set_certificate_file(cert_file)

        if ca_cert_file:
            Stream.ssl_set_ca_cert_file(ca_cert_file)

    def start(self):
        conn = connection.Connection(
            self, timeout=CONF.ovsdb_connection_timeout)
        ovsdbNbConn = OvsdbNbOvnIdl(conn)
        if self._events:
            self.notify_handler.watch_events(self._events)
        return ovsdbNbConn


class OvnSbIdl(OvnIdl):
    SCHEMA = 'OVN_Southbound'

    def __init__(self, connection_string, chassis=None, events=None,
                 tables=None):
        if connection_string.startswith("ssl"):
            self._check_and_set_ssl_files(self.SCHEMA)
        helper = self._get_ovsdb_helper(connection_string)
        self._events = events
        if tables is None:
            tables = ('Chassis', 'Encap', 'Port_Binding', 'Datapath_Binding',
                      'SB_Global')
        for table in tables:
            helper.register_table(table)
        super(OvnSbIdl, self).__init__(
            None, connection_string, helper, leader_only=False)
        if chassis:
            table = ('Chassis_Private' if 'Chassis_Private' in tables
                     else 'Chassis')
            self.tables[table].condition = [['name', '==', chassis]]

    def _get_ovsdb_helper(self, connection_string):
        return idlutils.get_schema_helper(connection_string, self.SCHEMA)

    def _check_and_set_ssl_files(self, schema_name):
        priv_key_file = CONF.ovn.ovn_sb_private_key
        cert_file = CONF.ovn.ovn_sb_certificate
        ca_cert_file = CONF.ovn.ovn_sb_ca_cert

        if priv_key_file:
            Stream.ssl_set_private_key_file(priv_key_file)

        if cert_file:
            Stream.ssl_set_certificate_file(cert_file)

        if ca_cert_file:
            Stream.ssl_set_ca_cert_file(ca_cert_file)

    def start(self):
        conn = connection.Connection(
            self, timeout=CONF.ovsdb_connection_timeout)
        ovsdbSbConn = OvsdbSbOvnIdl(conn)
        if self._events:
            self.notify_handler.watch_events(self._events)
        return ovsdbSbConn


class Backend(ovs_idl.Backend):
    lookup_table = {}
    ovsdb_connection = None

    def __init__(self, connection):
        self.ovsdb_connection = connection
        super(Backend, self).__init__(connection)

    @property
    def idl(self):
        return self.ovsdb_connection.idl

    @property
    def tables(self):
        return self.idl.tables


class StaticMACBindingFindCommand(command.DbFindCommand):
    table = 'Static_MAC_Binding'

    def __init__(self, api, port, ip):
        super().__init__(
            api,
            self.table,
            ('logical_port', '=', port),
            ('ip', '=', ip),
            row=True,
        )


# FIXME(ltomasbo): This can be removed once ovsdbapp supports it
class StaticMACBindingAddCommand(command.AddCommand):
    table_name = 'Static_MAC_Binding'

    def __init__(self, api, port, ip, mac, override_dynamic_mac=False,
                 may_exist=False, **columns):
        super().__init__(api)
        self.port = port
        self.ip = ip
        self.mac = mac
        self.override_dynamic_mac = override_dynamic_mac
        self.may_exist = may_exist
        self.columns = columns

    def run_idl(self, txn):
        cmd = StaticMACBindingFindCommand(self.api, self.port, self.ip)
        cmd.run_idl(txn)
        static_mac_binding_result = cmd.result
        if static_mac_binding_result:
            if len(static_mac_binding_result) > 1:
                # With the current database schema, this cannot happen, but
                # better safe than sorry.
                raise RuntimeError(
                    "Unexpected duplicates in database for port %s "
                    "and ip %s" % (self.port, self.ip))
            binding = static_mac_binding_result[0]
            if self.may_exist:
                # When no changes are made to a record, the parent
                # `post_commit` method will not be called.
                #
                # Ensure consistent return to caller of `Command.execute()`
                # even when no changes have been applied.
                self.result = rowview.RowView(binding)
                return
            else:
                raise RuntimeError(
                    "Static MAC Binding entry for port %s and ip %s exists" % (
                        self.port, self.ip))
        binding = txn.insert(self.api.tables[self.table_name])
        binding.logical_port = self.port
        binding.ip = self.ip
        binding.mac = self.mac
        binding.override_dynamic_mac = self.override_dynamic_mac
        self.set_columns(binding, **self.columns)
        # Setting the result to something other than a :class:`rowview.RowView`
        # or :class:`ovs.db.idl.Row` typed value will make the parent
        # `post_commit` method retrieve the newly insterted row from IDL and
        # return that to the caller.
        self.result = binding.uuid


# FIXME(ltomasbo): This can be removed once ovsdbapp supports it
class StaticMACBindingDelCommand(command.BaseCommand):
    table_name = 'Static_MAC_Binding'

    def __init__(self, api, port, ip, if_exists=False, **columns):
        super().__init__(api)
        self.port = port
        self.ip = ip
        self.if_exists = if_exists
        self.columns = columns

    def run_idl(self, txn):
        cmd = StaticMACBindingFindCommand(self.api, self.port, self.ip)
        cmd.run_idl(txn)
        static_mac_binding_result = cmd.result
        if static_mac_binding_result:
            if len(static_mac_binding_result) > 1:
                # With the current database schema, this cannot happen, but
                # better safe than sorry.
                raise RuntimeError(
                    "Unexpected duplicates in database for port %s "
                    "and ip %s" % (self.port, self.ip))
            binding = static_mac_binding_result[0]
            binding.delete()
            return
        if self.if_exists:
            return
        else:
            raise RuntimeError(
                "Static MAC Binding entry for port %s and ip %s does not "
                "exist" % (self.port, self.ip))


class GetLSPsForGwChassisCommand(command.ReadOnlyCommand):
    """Get all LSPs associated with a NAT entry.

    (FIPs) which its CR-LRP ports (gateway ports) are hosted on the given
    chassis.

    The resulted format is in tuple that can be consumed by the `wire` module.
    """
    def __init__(self, api, chassis_id):
        super().__init__(api)
        self.chassis_id = chassis_id

    def run_idl(self, txn):
        self.result = []
        for nat in self.api.tables['NAT'].rows.values():
            if nat.type != 'dnat_and_snat':
                continue

            try:
                gw_chassis = nat.gateway_port[0].status.get(
                    constants.OVN_STATUS_CHASSIS)
            except IndexError:
                continue

            if gw_chassis == self.chassis_id:
                ls_name = "neutron-{}".format(
                    nat.external_ids[constants.OVN_FIP_NET_EXT_ID_KEY])
                try:
                    lsp = self.api.lookup(
                        'Logical_Switch_Port', nat.logical_port[0])
                except IndexError:
                    LOG.debug("NAT %s does not have a logical_port set", nat)
                    continue
                self.result.append(
                    (nat.external_ip,
                     nat.gateway_port[0].mac,
                     ls_name,
                     lsp))


class OvsdbNbOvnIdl(nb_impl_idl.OvnNbApiIdlImpl, Backend):
    def __init__(self, connection):
        super(OvsdbNbOvnIdl, self).__init__(connection)
        self.idl._session.reconnect.set_probe_interval(60000)

    def get_network_vlan_tags(self):
        tags = []
        cmd = self.db_find_rows('Logical_Switch_Port', ('type', '=',
                                constants.OVN_LOCALNET_VIF_PORT_TYPE))
        for row in cmd.execute(check_error=True):
            if row.tag:
                tags.append(row.tag[0])
        return tags

    def get_network_vlan_tag_by_network_name(self, network_name):
        tags = []
        for row in self.get_localnet_ports_by_network_name(network_name):
            if row.tag:
                tags.append(row.tag[0])
        return tags

    def get_localnet_ports_by_network_name(self, network_name):
        conditions = (
            ('options', '=', {'network_name': network_name}),
            ('type', '=', constants.OVN_LOCALNET_VIF_PORT_TYPE),
        )
        cmd = self.db_find_rows('Logical_Switch_Port', *conditions)
        for row in cmd.execute(check_error=True):
            yield row

    def get_bgpvpn_networks_for_ports(self, ports,
                                      vpn_type=constants.OVN_EVPN_TYPE_L3):
        cmd = self.db_find_rows(
            'Logical_Switch',
            ('external_ids', '=', {constants.OVN_EVPN_TYPE_EXT_ID_KEY:
                                   vpn_type}))
        networks = []
        for row in cmd.execute(check_error=True):
            if not row.ports:
                continue

            localnet_ports = [p for p in row.ports if p in ports]
            if localnet_ports:
                networks.append(row)
        return networks

    def ls_has_virtual_ports(self, logical_switch):
        ls = self.lookup('Logical_Switch', logical_switch)
        for port in ls.ports:
            if port.type == constants.OVN_VIRTUAL_VIF_PORT_TYPE:
                return True
        return False

    def get_nat_by_logical_port(self, logical_port):
        cmd = self.db_find_rows('NAT', ('logical_port', '=', logical_port))
        nat_info = cmd.execute(check_error=True)
        return nat_info[0] if nat_info else []

    def get_active_lsp_on_chassis(self, chassis):
        ports = []
        cmd = self.db_find_rows('Logical_Switch_Port', ('up', '=', True))
        for row in cmd.execute(check_error=True):
            port_chassis = driver_utils.get_port_chassis(row, chassis)
            if port_chassis == chassis:
                ports.append(row)

        return ports

    def get_active_cr_lrp_on_chassis(self, chassis):
        ports = []
        rows = self.db_list_rows('Logical_Router_Port').execute(
            check_error=True)

        for row in rows:
            if not hasattr(row, 'status'):
                LOG.warning("OVN version does not include support for status "
                            "information. Therefore router ports and tenant "
                            "IPs cannot be exposed.")
                break
            if row.status.get(constants.OVN_STATUS_CHASSIS) == chassis:
                ports.append(row)
        return ports

    def get_active_local_lrps(self, local_gateway_ports):
        ports = []
        cmd = self.db_find_rows(
            'Logical_Switch_Port', ('up', '=', True),
            ('type', '=', constants.OVN_ROUTER_PORT_TYPE),
            ('external_ids', '=', {constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                                   constants.OVN_ROUTER_INTERFACE}))
        for row in cmd.execute(check_error=True):
            if (row.external_ids.get(constants.OVN_DEVICE_ID_EXT_ID_KEY)
                    in local_gateway_ports):
                ports.append(row)
        return ports

    def get_active_lsp(self, network):
        ports = []
        # port type ""
        cmd = self.db_find_rows(
            'Logical_Switch_Port', ('up', '=', True),
            ('type', '=', constants.OVN_VM_VIF_PORT_TYPE),
            ('external_ids', '=', {constants.OVN_LS_NAME_EXT_ID_KEY: network}))
        ports.extend(cmd.execute(check_error=True))
        # port type "virtual"
        cmd = self.db_find_rows(
            'Logical_Switch_Port', ('up', '=', True),
            ('type', '=', constants.OVN_VIRTUAL_VIF_PORT_TYPE),
            ('external_ids', '=', {constants.OVN_LS_NAME_EXT_ID_KEY: network}))
        ports.extend(cmd.execute(check_error=True))
        return ports

    def get_active_local_lbs(self, local_gateway_ports):
        lbs = []
        cmd = self.db_find_rows('Load_Balancer', ('vips', '!=', {}))

        for row in cmd.execute(check_error=True):
            for ext_id_key in constants.OVN_LB_EXT_ID_ROUTER_KEY:
                try:
                    router_name = row.external_ids[ext_id_key].replace(
                        'neutron-', '', 1)
                    if router_name in local_gateway_ports:
                        lbs.append(row)
                        break
                except KeyError:
                    continue
        return lbs

    def static_mac_binding_add(self, port, ip, mac, override_dynamic_mac=False,
                               may_exist=False):
        return StaticMACBindingAddCommand(self, port, ip, mac,
                                          override_dynamic_mac, may_exist)

    def static_mac_binding_del(self, port, ip, if_exists=False):
        return StaticMACBindingDelCommand(self, port, ip, if_exists)

    def get_router(self, router):
        router_name = 'neutron-' + router
        return self.lr_get(router_name).execute(check_error=True)

    def get_distributed_flag(self):
        """Return distributed flag set by Neutron.

        The method returns True by default if the flag is not set in Neutron.
        """
        nb_global_ext_ids = self.db_get(
            'NB_Global', '.', 'external_ids').execute(check_error=True)
        try:
            distributed = nb_global_ext_ids[constants.OVN_FIP_DISTRIBUTED]
        except KeyError:
            LOG.warning(
                "Agent did not find %s configuration in the OVN Northbound "
                "NB_Global table. This option determines if the traffic is "
                "centralized or distributed. The  agent default to "
                "distributed traffic but be aware Neutron by  default works "
                "in centralized manner. Please make sure Neutron has "
                "the enable_distributed_floating_ip config option set "
                "to True." % constants.OVN_FIP_DISTRIBUTED)
            distributed = "True"
        return True if distributed == "True" else False

    def get_nats_by_lrp(self, lrp):
        return self.db_find_rows(
            'NAT',
            ('gateway_port', '=', lrp.uuid),
            ('type', '=', constants.OVN_DNAT_AND_SNAT)
        ).execute(check_error=True)


class OvsdbSbOvnIdl(sb_impl_idl.OvnSbApiIdlImpl, Backend):
    def __init__(self, connection):
        super(OvsdbSbOvnIdl, self).__init__(connection)
        self.idl._session.reconnect.set_probe_interval(60000)

    def get_port_by_name(self, port):
        cmd = self.db_find_rows('Port_Binding', ('logical_port', '=', port))
        port_info = cmd.execute(check_error=True)
        return port_info[0] if port_info else []

    def get_ports_on_datapath(self, datapath, port_type=None):
        if port_type:
            cmd = self.db_find_rows('Port_Binding',
                                    ('datapath', '=', datapath),
                                    ('type', '=', port_type))
        else:
            cmd = self.db_find_rows('Port_Binding',
                                    ('datapath', '=', datapath))
        try:
            return cmd.execute(check_error=True)
        except ValueError:
            # Datapath has been removed.
            raise exceptions.DatapathNotFound(datapath=datapath)

    def get_ports_by_type(self, port_type):
        cmd = self.db_find_rows('Port_Binding',
                                ('type', '=', port_type))
        return cmd.execute(check_error=True)

    def is_provider_network(self, datapath):
        try:
            cmd = self.db_find_rows('Port_Binding',
                                    ('datapath', '=', datapath),
                                    ('type', '=',
                                     constants.OVN_LOCALNET_VIF_PORT_TYPE))
            return bool(cmd.execute(check_error=True))
        except ValueError:
            # Datapath has been removed.
            raise exceptions.DatapathNotFound(datapath=datapath)

    def get_localnet_for_datapath(self, datapath):
        try:
            cmd = self.db_find_rows('Port_Binding',
                                    ('datapath', '=', datapath),
                                    ('type', '=',
                                     constants.OVN_LOCALNET_VIF_PORT_TYPE))
            localnet_info = cmd.execute(check_error=True)
            return localnet_info[0].logical_port if localnet_info else []
        except ValueError:
            # Datapath has been removed.
            raise exceptions.DatapathNotFound(datapath=datapath)

    def get_fip_associated(self, port):
        cmd = self.db_find_rows(
            'Port_Binding', ('type', '=', constants.OVN_PATCH_VIF_PORT_TYPE))
        for row in cmd.execute(check_error=True):
            for fip in row.nat_addresses:
                if port in fip:
                    return fip.split(" ")[1], row.datapath
        return None, None

    def is_port_on_chassis(self, port_name, chassis):
        port_info = self.get_port_by_name(port_name)
        try:
            return (port_info and
                    port_info.chassis[0].name == chassis)
        except IndexError:
            pass
        return False

    def is_port_without_chassis(self, port_name):
        port_info = self.get_port_by_name(port_name)
        return (port_info and not port_info.chassis)

    def is_port_deleted(self, port_name):
        return False if self.get_port_by_name(port_name) else True

    def get_ports_on_chassis(self, chassis):
        rows = self.db_list_rows('Port_Binding').execute(check_error=True)
        return [r for r in rows if r.chassis and r.chassis[0].name == chassis]

    def get_cr_lrp_ports(self):
        return self.db_find_rows(
            "Port_Binding",
            ("type", "=", constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE),
        ).execute(check_error=True)

    def get_cr_lrp_ports_on_chassis(self, chassis):
        return [
            r.logical_port
            for r in self.get_cr_lrp_ports()
            if r.chassis and r.chassis[0].name == chassis
        ]

    def get_cr_lrp_nat_addresses_info(self, cr_lrp_port_name, chassis, sb_idl):
        # NOTE: Assuming logical_port format is "cr-lrp-XXXX"
        patch_port_name = cr_lrp_port_name.split("cr-lrp-")[1]
        patch_port_row = self.get_port_by_name(patch_port_name)
        if not patch_port_row:
            return [], None
        ips = []
        for row in patch_port_row.nat_addresses:
            nat_ips = row.split(" ")[1:-1]
            port = row.split(" ")[-1].split("\"")[1]
            if port and sb_idl and sb_idl.is_port_on_chassis(port, chassis):
                ips.extend(nat_ips)
        return ips, patch_port_row

    def get_provider_datapath_from_cr_lrp(self, cr_lrp):
        if cr_lrp.startswith('cr-lrp'):
            provider_port = cr_lrp.split("cr-lrp-")[1]
            return self.get_port_datapath(provider_port)
        return None

    def get_datapath_from_port_peer(self, port):
        peer_name = port.options['peer']
        return self.get_port_datapath(peer_name)

    def get_network_name_and_tag(self, datapath, bridge_mappings):
        for row in self.get_ports_on_datapath(
                datapath, constants.OVN_LOCALNET_VIF_PORT_TYPE):
            if (row.options and
                    row.options.get('network_name') in bridge_mappings):
                return row.options.get('network_name'), row.tag
        return None, None

    def get_network_vlan_tags(self):
        tags = []
        cmd = self.db_find_rows('Port_Binding', ('type', '=',
                                constants.OVN_LOCALNET_VIF_PORT_TYPE))
        for row in cmd.execute(check_error=True):
            if row.tag:
                tags.append(row.tag[0])
        return tags

    def get_network_vlan_tag_by_network_name(self, network_name):
        tags = []
        cmd = self.db_find_rows('Port_Binding', ('type', '=',
                                constants.OVN_LOCALNET_VIF_PORT_TYPE))
        for row in cmd.execute(check_error=True):
            if (row.tag and row.options and
                    row.options.get('network_name') == network_name):
                tags.append(row.tag[0])
        return tags

    def is_router_gateway_on_chassis(self, datapath, chassis):
        port_info = self.get_ports_on_datapath(
            datapath, constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        try:
            if port_info and port_info[0].chassis[0].name == chassis:
                return port_info[0].logical_port
        except IndexError:
            return False

    def is_router_gateway_on_any_chassis(self, datapath):
        port_info = self.get_ports_on_datapath(
            datapath, constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        try:
            if port_info and port_info[0].chassis[0].name:
                return port_info[0]
        except IndexError:
            return False

    def get_lrps_for_datapath(self, datapath):
        lrps = []
        for row in self.get_ports_on_datapath(
                datapath, constants.OVN_PATCH_VIF_PORT_TYPE):
            if row.options:
                lrps.append(row.options['peer'])
        return lrps

    def get_lrp_ports_for_router(self, datapath):
        return self.get_ports_on_datapath(
            datapath, constants.OVN_PATCH_VIF_PORT_TYPE)

    def get_lrp_ports_on_provider(self):
        provider_lrp_ports = []
        lrp_ports = self.get_ports_by_type(constants.OVN_PATCH_VIF_PORT_TYPE)
        for lrp_port in lrp_ports:
            if lrp_port.logical_port.startswith(
                    constants.OVN_LRP_PORT_NAME_PREFIX):
                continue
            if self.is_provider_network(lrp_port.datapath):
                provider_lrp_ports.append(lrp_port)

    def get_port_datapath(self, port_name):
        port_info = self.get_port_by_name(port_name)
        if port_info:
            return port_info.datapath

    def get_ip_from_port_peer(self, port):
        peer_name = port.options['peer']
        peer_port = self.get_port_by_name(peer_name)
        try:
            return peer_port.mac[0].split(' ')[1]
        except AttributeError:
            raise exceptions.PortNotFound(port=peer_name)

    def get_evpn_info_from_port_name(self, port_name):
        if port_name.startswith(constants.OVN_CRLRP_PORT_NAME_PREFIX):
            port_name = port_name.split(
                constants.OVN_CRLRP_PORT_NAME_PREFIX)[1]
        elif port_name.startswith(constants.OVN_LRP_PORT_NAME_PREFIX):
            port_name = port_name.split(constants.OVN_LRP_PORT_NAME_PREFIX)[1]

        port = self.get_port_by_name(port_name)
        return self.get_evpn_info(port)

    def get_evpn_info(self, port):
        try:
            return {'vni': int(
                    port.external_ids[constants.OVN_EVPN_VNI_EXT_ID_KEY]),
                    'bgp_as': int(
                    port.external_ids[constants.OVN_EVPN_AS_EXT_ID_KEY])}
        except (KeyError, ValueError):
            LOG.debug('Either "%s" or "%s" were not found or have an '
                      'invalid value in the port %s '
                      'external_ids %s', constants.OVN_EVPN_VNI_EXT_ID_KEY,
                      constants.OVN_EVPN_AS_EXT_ID_KEY, port.logical_port,
                      port.external_ids)
            return {}

    def get_port_if_local_chassis(self, port_name, chassis):
        port = self.get_port_by_name(port_name)
        if port.chassis[0].name == chassis:
            return port

    def get_virtual_ports_on_datapath_by_chassis(self, datapath, chassis):
        try:
            rows = self.get_ports_on_datapath(
                datapath, port_type=constants.OVN_VIRTUAL_VIF_PORT_TYPE)
        except exceptions.DatapathNotFound:
            return []
        return [r for r in rows if r.chassis and r.chassis[0].name == chassis]

    def get_ovn_lb(self, name):
        cmd = self.db_find_rows('Load_Balancer', ('name', '=', name))
        lb_info = cmd.execute(check_error=True)
        return lb_info[0] if lb_info else []

    def get_provider_ovn_lbs_on_cr_lrp(self, provider_dp, router_dp):
        # return {vip_port: vip_ip, vip_port2: vip_ip2, ...}
        # ovn-sbctl find port_binding type=\"\" chassis=[] mac=[] up=false
        cmd = self.db_find_rows('Port_Binding',
                                ('datapath', '=', provider_dp),
                                ('type', '=', constants.OVN_VM_VIF_PORT_TYPE),
                                ('chassis', '=', []),
                                ('mac', '=', []),
                                ('up', '=', False))
        lbs = {}
        for row in cmd.execute(check_error=True):
            # This is depending on the external-id information added by
            # neutron, regarding the neutron:cidrs
            ip_info = row.external_ids.get(
                constants.OVN_CIDRS_EXT_ID_KEY, "")
            if not ip_info:
                continue
            port_name = row.external_ids.get(
                constants.OVN_PORT_NAME_EXT_ID_KEY)
            if (not port_name or
                    len(port_name) <= len(constants.LB_VIP_PORT_PREFIX)):
                continue
            lb_name = port_name[len(constants.LB_VIP_PORT_PREFIX):]
            lb = self.get_ovn_lb(lb_name)
            if not lb:
                continue
            lb_dps, lr_dps = helpers.get_lb_datapaths(lb)
            if not lb_dps:
                lb_dps = lb.datapaths

            if not lr_dps:
                # assume all the members are connected through the same router
                # so only one datapath needs to be checked
                router_lrps = self.get_lrps_for_datapath(lb_dps[0])
                for lrp in router_lrps:
                    router_lrp_dp = self.get_port_datapath(lrp)
                    if router_lrp_dp == router_dp:
                        lb_ip = ip_info.split(" ")[0].split("/")[0]
                        lbs[lb.name] = lb_ip
                        break
            elif router_dp in lr_dps:
                lb_ip = ip_info.split(" ")[0].split("/")[0]
                lbs[lb.name] = lb_ip

        return lbs

    def get_ovn_vip_port(self, name):
        # ovn-sbctl find port_binding type=\"\" chassis=[] mac=[] up=false
        cmd = self.db_find_rows('Port_Binding',
                                ('type', '=', constants.OVN_VM_VIF_PORT_TYPE),
                                ('chassis', '=', []),
                                ('mac', '=', []),
                                ('up', '=', False))

        for row in cmd.execute(check_error=True):
            port_name = row.external_ids.get(
                constants.OVN_PORT_NAME_EXT_ID_KEY)
            if port_name[len(constants.LB_VIP_PORT_PREFIX):] == name:
                return row



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import ovn_evpn_driver
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests.unit import fakes
from ovn_bgp_agent.utils import linux_net

CONF = cfg.CONF


class TestOVNEVPNDriver(test_base.TestCase):

    def setUp(self):
        super(TestOVNEVPNDriver, self).setUp()
        self.evpn_driver = ovn_evpn_driver.OVNEVPNDriver()
        self.mock_sbdb = mock.patch.object(ovn, 'OvnSbIdl').start()
        self.mock_ovs_idl = mock.patch.object(ovs, 'OvsIdl').start()
        self.evpn_driver.ovs_idl = self.mock_ovs_idl
        self.evpn_driver.sb_idl = mock.Mock()
        self.sb_idl = self.evpn_driver.sb_idl
        self.evpn_driver.chassis = 'fake-chassis'
        self.ipv4 = '192.168.1.17'
        self.ipv6 = '2002::1234:abcd:ffff:c0a8:101'
        self.fip = '172.24.4.33'
        self.mac = 'aa:bb:cc:dd:ee:ff'
        self.mac1 = 'aa:bb:cc:dd:ee:ee'
        self.bridge = 'fake-bridge'
        self.vni = 77
        self.vni1 = 88
        self.vlan_tag = 10
        self.evpn_driver.ovn_bridge_mappings = {'fake-network': self.bridge}
        self.evpn_info = {'bgp_as': 'fake-bgp-as', 'vni': self.vni}
        self.evpn_device = fakes.create_object({
            'lo_name': 'fake-lo-name',
            'bridge_name': self.bridge,
            'vxlan_name': 'fake-vxlan-name',
            'vrf_name': 'fake-vrf-name',
            'veth_vrf': 'fake-veth-vrf',
            'veth_ovs': 'fake-veth-ovs',
            'vlan_name': 'fake-vlan-name'})
        self.cr_lrp = 'cr-fake-logical-port'
        self.cr_lrp1 = 'cr-fake-logical-port1'
        self.evpn_driver.ovn_local_cr_lrps = {
            self.cr_lrp: {
                'provider_datapath': 'fake-provider-dp',
                'ips': [self.fip],
                'vni': self.vni,
                'bgp_as': 'fake-bgp-as',
                'bridge': self.bridge,
                'vlan': 'fake-vlan',
                'vxlan': 'fake-vxlan',
                'vrf': 'fake-vrf',
                'veth_vrf': 'fake-veth-vrf',
                'veth_ovs': 'fake-veth-ovs',
                'lo': 'fake-lo',
                'mac': self.mac},
            self.cr_lrp1: {
                'provider_datapath': 'fake-provider-dp1',
                'ips': [self.fip],
                'vni': self.vni1,
                'bgp_as': 'fake-bgp-as1',
                'bridge': self.bridge,
                'vlan': 'fake-vlan1',
                'vxlan': 'fake-vxlan1',
                'vrf': 'fake-vrf1',
                'veth_vrf': 'fake-veth-vrf1',
                'veth_ovs': 'fake-veth-ovs1',
                'lo': 'fake-lo1',
                'mac': self.mac1},
        }
        self.evpn_driver._ovn_routing_tables_routes = {
            'fake-vlan': [{
                'route': {
                    'oif': 'fake-oif',
                    'gateway': 'fake-gateway',
                    'dst': '{}/32'.format(self.ipv4),
                    'dst_len': 32,
                    'table': 'fake-table'},
                'vlan': 88,
                }]
            }

    def test_start(self):
        self.evpn_driver.start()
        # Assert connections were started
        self.mock_ovs_idl().start.assert_called_once_with(
            CONF.ovsdb_connection)
        self.mock_sbdb().start.assert_called_once_with()

    @mock.patch.object(linux_net, 'ensure_arp_ndp_enabled_for_bridge')
    def test_sync(self, mock_ensure_ndp):
        self.mock_ovs_idl.get_ovn_bridge_mappings.return_value = [
            'net0:bridge0', 'net1:bridge1']
        port0 = fakes.create_object({
            'name': 'fake-port0',
            'type': constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE})
        port1 = fakes.create_object({
            'name': 'fake-port1',
            'type': constants.OVN_VIRTUAL_VIF_PORT_TYPE})
        self.sb_idl.get_ports_on_chassis.return_value = [
            port0, port1]

        mock_expose_ip = mock.patch.object(
            self.evpn_driver, '_expose_ip').start()
        mock_remove_extra_exposed_ips = mock.patch.object(
            self.evpn_driver, '_remove_extra_exposed_ips').start()
        mock_remove_extra_routes = mock.patch.object(
            self.evpn_driver, '_remove_extra_routes').start()
        mock_remove_extra_ovs_flows = mock.patch.object(
            self.evpn_driver, '_remove_extra_ovs_flows').start()
        mock_remove_extra_vrfs = mock.patch.object(
            self.evpn_driver, '_remove_extra_vrfs').start()

        self.evpn_driver.sync()

        expected_calls = [mock.call('bridge0', 1), mock.call('bridge1', 2)]
        mock_ensure_ndp.assert_has_calls(expected_calls)
        mock_expose_ip.assert_called_once_with(port0, cr_lrp=True)
        mock_remove_extra_exposed_ips.assert_called_once_with()
        mock_remove_extra_routes.assert_called_once_with()
        mock_remove_extra_ovs_flows.assert_called_once_with()
        mock_remove_extra_vrfs.assert_called_once_with()

    def test__ensure_network_exposed(self):
        self.sb_idl.get_evpn_info_from_port_name.return_value = 'fake-info'
        self.sb_idl.get_port_datapath.return_value = 'fake-dp'
        mock_expose_subnet = mock.patch.object(
            self.evpn_driver, '_expose_subnet').start()
        mock_get_bridge = mock.patch.object(
            self.evpn_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, self.vlan_tag)
        gateway = {}
        gateway['ips'] = ['10.10.10.1/32']
        gateway['provider_datapath'] = 'fake-prov-dp'
        lrp = fakes.create_object({
            'name': 'fake-lrp',
            'logical_port': self.cr_lrp,
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)],
            'options': {'peer': 'fake-peer'},
            'datapath': 'fake-dp'})

        self.evpn_driver._ensure_network_exposed(lrp, gateway)

        mock_expose_subnet.assert_called_once_with(
            self.ipv4, ['10.10.10.1'],
            {'ips': ['10.10.10.1/32'], 'provider_datapath': 'fake-prov-dp'},
            self.bridge, self.vlan_tag, 'fake-dp')

    def test__get_bridge_for_datapath(self):
        self.sb_idl.get_network_name_and_tag.return_value = (
            'fake-network', [self.vlan_tag])
        ret = self.evpn_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((self.bridge, self.vlan_tag), ret)

    def test__get_bridge_for_datapath_no_tag(self):
        self.sb_idl.get_network_name_and_tag.return_value = (
            'fake-network', None)
        ret = self.evpn_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((self.bridge, None), ret)

    def test__get_bridge_for_datapath_no_network_name(self):
        self.sb_idl.get_network_name_and_tag.return_value = (None, None)
        ret = self.evpn_driver._get_bridge_for_datapath('fake-dp')
        self.assertEqual((None, None), ret)

    @mock.patch.object(linux_net, 'add_ip_nei')
    @mock.patch.object(frr, 'vrf_reconfigure')
    def _test_expose_ip(
            self, mock_vrf_reconfigure, mock_add_ip_nei, cr_lrp=False):
        mock_get_bridge = mock.patch.object(
            self.evpn_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, self.vlan_tag)
        mock_ensure_evpn = mock.patch.object(
            self.evpn_driver, '_ensure_evpn_devices').start()
        mock_ensure_evpn.return_value = self.evpn_device
        mock_connect_evpn = mock.patch.object(
            self.evpn_driver, '_connect_evpn_to_ovn').start()
        mock_ensure_net_exposed = mock.patch.object(
            self.evpn_driver, '_ensure_network_exposed').start()
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': self.cr_lrp,
            'mac': ['{} {} {}'.format(self.mac, self.ipv4, self.ipv6)],
            'datapath': 'fake-dp'})
        self.sb_idl.get_fip_associated.return_value = (
            self.fip, 'fake-dp')
        lrp0 = fakes.create_object({'chassis': 'fake-chassis', 'options': {}})
        lrp1 = fakes.create_object({'chassis': '', 'options': {}})
        self.sb_idl.get_lrp_ports_for_router.return_value = [lrp0, lrp1]

        if not cr_lrp:
            self.sb_idl.get_port_if_local_chassis.return_value = row
            self.sb_idl.get_evpn_info.return_value = self.evpn_info
        else:
            self.sb_idl.get_evpn_info_from_port_name.return_value = (
                self.evpn_info)

        self.evpn_driver.expose_ip(row, cr_lrp=cr_lrp)

        # Assertions
        mock_connect_evpn.assert_called_once_with(
            'fake-vrf-name', 'fake-veth-vrf', 'fake-veth-ovs',
            [self.ipv4, self.ipv6], self.bridge, self.vni,
            'fake-vlan-name', self.vlan_tag)
        mock_ensure_evpn.assert_called_once_with(
            self.bridge, self.vni, self.vlan_tag)
        mock_ensure_net_exposed.assert_called_once_with(
            lrp1,
            {'router_datapath': 'fake-dp', 'provider_datapath': 'fake-dp',
             'ips': [self.ipv4, self.ipv6], 'mac': self.mac, 'vni': self.vni,
             'bgp_as': 'fake-bgp-as', 'lo': 'fake-lo-name',
             'bridge': self.bridge, 'vxlan': 'fake-vxlan-name',
             'vrf': 'fake-vrf-name', 'veth_vrf': 'fake-veth-vrf',
             'veth_ovs': 'fake-veth-ovs', 'vlan': 'fake-vlan-name'})
        mock_vrf_reconfigure.assert_called_once_with(
            self.evpn_info, action='add-vrf')
        expected_calls = [mock.call(self.ipv4, self.mac, 'fake-vlan-name'),
                          mock.call(self.ipv6, self.mac, 'fake-vlan-name')]
        mock_add_ip_nei.assert_has_calls(expected_calls)

    def test_expose_ip(self):
        self._test_expose_ip(cr_lrp=False)

    def test_expose_ip_cr_lrp(self):
        self._test_expose_ip(cr_lrp=True)

    @mock.patch.object(ovs, 'remove_evpn_router_ovs_flows')
    @mock.patch.object(frr, 'vrf_reconfigure')
    def _test_withdraw_ip(
            self, mock_vrf_reconfigure, mock_remove_evpn_flows, cr_lrp=True,
            ret_vlan_tag=True):
        mock_remove_evpn = mock.patch.object(
            self.evpn_driver, '_remove_evpn_devices').start()
        mock_disconnect_epvn = mock.patch.object(
            self.evpn_driver, '_disconnect_evpn_from_ovn').start()
        mock_get_bridge = mock.patch.object(
            self.evpn_driver, '_get_bridge_for_datapath').start()
        vlan_tag = self.vlan_tag if ret_vlan_tag else None
        mock_get_bridge.return_value = (self.bridge, vlan_tag)
        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': self.cr_lrp})

        self.evpn_driver.withdraw_ip(row, cr_lrp=cr_lrp)

        mock_remove_evpn_flows.assert_called_once_with(
            self.bridge, constants.OVS_VRF_RULE_COOKIE, self.mac)
        mock_vrf_reconfigure.assert_called_once_with(
            {'vni': self.vni, 'bgp_as': 'fake-bgp-as'}, action='del-vrf')

        kwargs = {}
        if ret_vlan_tag:
            kwargs.update({'vlan_tag': vlan_tag})
        mock_disconnect_epvn.assert_called_once_with(
            self.vni, self.bridge, [self.fip], **kwargs)

        mock_remove_evpn.assert_called_once_with(self.vni)

    def test_withdraw_ip(self):
        self._test_withdraw_ip()

    def test_withdraw_ip_no_vlan_tag(self):
        self._test_withdraw_ip(ret_vlan_tag=False)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        self.sb_idl.get_evpn_info_from_port_name.return_value = self.evpn_info
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.evpn_driver.ovn_local_lrps = {lrp: 'fake-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.expose_remote_ip(ips, row)

        lo_name = constants.OVN_EVPN_LO_PREFIX + str(self.vni)
        mock_add_ip_dev.assert_called_once_with(
            lo_name, ips, clear_local_route_at_table=self.vni)

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip_is_provider_network(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = True
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    def test_expose_remote_ip_not_local(self, mock_add_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.evpn_driver.ovn_local_lrps = {}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.expose_remote_ip(ips, row)

        mock_add_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        self.sb_idl.get_evpn_info_from_port_name.return_value = self.evpn_info
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.evpn_driver.ovn_local_lrps = {lrp: 'fake-lrp'}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.withdraw_remote_ip(ips, row)

        lo_name = constants.OVN_EVPN_LO_PREFIX + str(self.vni)
        mock_del_ip_dev.assert_called_once_with(lo_name, ips)

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip_is_provider_network(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = True
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_not_called()

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    def test_withdraw_remote_ip_not_local(self, mock_del_ip_dev):
        self.sb_idl.is_provider_network.return_value = False
        lrp = 'fake-lrp'
        self.sb_idl.get_lrps_for_datapath.return_value = [lrp]
        self.evpn_driver.ovn_local_lrps = {}
        row = fakes.create_object({
            'name': 'fake-row', 'datapath': 'fake-dp'})

        ips = [self.ipv4, self.ipv6]
        self.evpn_driver.withdraw_remote_ip(ips, row)

        mock_del_ip_dev.assert_not_called()

    def test_expose_subnet(self):
        self.sb_idl.get_evpn_info.return_value = self.evpn_info
        self.sb_idl.get_ip_from_port_peer.return_value = self.ipv4
        self.sb_idl.get_port_datapath.return_value = 'fake-dp'
        self.sb_idl.is_router_gateway_on_chassis.return_value = self.cr_lrp
        mock_get_bridge = mock.patch.object(
            self.evpn_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, self.vlan_tag)
        mock_expose_subnet = mock.patch.object(
            self.evpn_driver, '_expose_subnet').start()

        row = fakes.create_object({
            'name': 'fake-row',
            'datapath': 'fake-dp',
            'logical_port': self.cr_lrp})

        self.evpn_driver.expose_subnet(row)

        mock_expose_subnet.assert_called_once_with(
            self.ipv4, [self.fip],
            self.evpn_driver.ovn_local_cr_lrps[self.cr_lrp],
            self.bridge, self.vlan_tag, 'fake-dp')

    @mock.patch.object(linux_net, 'add_ips_to_dev')
    @mock.patch.object(ovs, 'ensure_evpn_ovs_flow')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    def _test__expose_subnet(
            self, mock_ip_version, mock_add_route, mock_ensure_evpn_flow,
            mock_add_ip_dev, use_ipv6=False):
        # IPv4 vs IPv6 mocks
        ip = self.ipv6 if use_ipv6 else self.ipv4
        mock_ip_version.return_value = (
            constants.IP_VERSION_6 if use_ipv6 else constants.IP_VERSION_4)
        cidr = '128' if use_ipv6 else '32'

        port0 = fakes.create_object({
            'name': 'fake-port0',
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_port': self.cr_lrp,
            'chassis': 'fake-chassis',
            'mac': ['{} {}'.format(self.mac, ip)],
            'datapath': 'fake-dp'})
        port1 = fakes.create_object({
            'name': 'fake-port1',
            'chassis': 'fake-chassis',
            'type': 'unknown-type'})
        self.sb_idl.get_ports_on_datapath.return_value = [port0, port1]

        self.evpn_driver._expose_subnet(
            '{}/{}'.format(ip, cidr), [self.fip],
            self.evpn_driver.ovn_local_cr_lrps[self.cr_lrp], self.bridge,
            10, 'fake-dp')

        mock_add_route.assert_called_once_with(
            mock.ANY, ip, self.vni, 'fake-vlan',
            mask=cidr, via=self.fip)
        mock_ensure_evpn_flow.assert_called_once_with(
            self.bridge, constants.OVS_VRF_RULE_COOKIE, self.mac,
            'fake-vlan', 'fake-vlan', '{}/{}'.format(ip, cidr),
            strip_vlan=True)
        mock_add_ip_dev.assert_called_once_with(
            'fake-lo', [ip], clear_local_route_at_table=self.vni)

    def test__expose_subnet(self):
        self._test__expose_subnet()

    def test__expose_subnet_ipv6(self):
        self._test__expose_subnet(use_ipv6=True)

    @mock.patch.object(linux_net, 'delete_exposed_ips')
    @mock.patch.object(linux_net, 'get_exposed_ips_on_network')
    @mock.patch.object(ovs, 'remove_evpn_network_ovs_flow')
    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    def _test_withdraw_subnet(
            self, mock_ip_version, mock_del_route, mock_remove_evpn_flows,
            mock_get_ips, mock_del_ips, use_ipv6=False):
        # IPv4 vs IPv6 mocks
        ip = self.ipv6 if use_ipv6 else self.ipv4
        mock_ip_version.return_value = (
            constants.IP_VERSION_6 if use_ipv6 else constants.IP_VERSION_4)
        cidr = '128' if use_ipv6 else '32'

        self.evpn_driver.ovn_local_lrps = {
            'lrp-port': {'datapath': 'fake-dp',
                         'ip': '{}/{}'.format(ip, cidr)}}
        self.sb_idl.is_router_gateway_on_chassis.return_value = self.cr_lrp
        mock_get_bridge = mock.patch.object(
            self.evpn_driver, '_get_bridge_for_datapath').start()
        mock_get_bridge.return_value = (self.bridge, self.vlan_tag)
        net_ips = ['10.10.10.1', '2001:0db8:85a3:0000:0000:8a2e:0370:7334']
        mock_get_ips.return_value = net_ips

        row = fakes.create_object({
            'name': 'fake-row',
            'logical_port': 'port'})
        self.evpn_driver.withdraw_subnet(row)

        mock_del_route.assert_called_once_with(
            mock.ANY, ip, self.vni, 'fake-vlan',
            mask=cidr, via=self.fip)
        mock_remove_evpn_flows.assert_called_once_with(
            self.bridge, constants.OVS_VRF_RULE_COOKIE, self.mac,
            '{}/{}'.format(ip, cidr))
        mock_del_ips.assert_called_once_with(net_ips, 'fake-lo')

    def test_withdraw_subnet(self):
        self._test_withdraw_subnet()

    def test_withdraw_subnet_ipv6(self):
        self._test_withdraw_subnet(use_ipv6=True)

    @mock.patch.object(linux_net, 'ensure_veth')
    @mock.patch.object(linux_net, 'enable_proxy_ndp')
    @mock.patch.object(linux_net, 'set_device_status')
    @mock.patch.object(ovs, 'add_vlan_port_to_ovs_bridge')
    @mock.patch.object(linux_net, 'get_nic_ip')
    @mock.patch.object(linux_net, 'ensure_dummy_device')
    @mock.patch.object(linux_net, 'ensure_vxlan')
    @mock.patch.object(linux_net, 'set_master_for_device')
    @mock.patch.object(linux_net, 'ensure_bridge')
    @mock.patch.object(linux_net, 'ensure_vrf')
    def _test__ensure_evpn_devices(
            self, mock_ensure_vrf, mock_ensure_bridge, mock_set_master,
            mock_ensure_vxlan, mock_ensure_dummy, mock_get_nic_ip,
            mock_add_vlan, mock_set_status, mock_proxy_ndp,
            mock_ensure_veth, use_vlan=True):

        mock_get_nic_ip.return_value = [self.ipv4]

        dp_bridge = 'datapath-bridge'
        vlan_tag = self.vlan_tag if use_vlan else None
        self.evpn_driver._ensure_evpn_devices(dp_bridge, self.vni, vlan_tag)

        # Asserts
        vrf_name = constants.OVN_EVPN_VRF_PREFIX + str(self.vni)
        mock_ensure_vrf.assert_called_once_with(vrf_name, self.vni)

        bridge_name = constants.OVN_EVPN_BRIDGE_PREFIX + str(self.vni)
        mock_ensure_bridge.assert_called_once_with(bridge_name)

        vxlan_name = constants.OVN_EVPN_VXLAN_PREFIX + str(self.vni)
        mock_ensure_vxlan.assert_called_once_with(
            vxlan_name, self.vni, self.ipv4, CONF.evpn_udp_dstport)

        lo_name = constants.OVN_EVPN_LO_PREFIX + str(self.vni)
        mock_ensure_dummy.assert_called_once_with(lo_name)

        set_master_expected_calls = [
            mock.call(bridge_name, vrf_name),
            mock.call(vxlan_name, bridge_name),
            mock.call(lo_name, vrf_name)]

        if use_vlan:
            vlan_name = constants.OVN_EVPN_VLAN_PREFIX + str(self.vni)
            mock_add_vlan.assert_called_once_with(
                dp_bridge, vlan_name, self.vlan_tag)

            mock_set_status.assert_called_once_with(
                vlan_name, constants.LINK_UP)
            mock_proxy_ndp.assert_called_once_with(vlan_name)

            set_master_expected_calls.append(mock.call(vlan_name, vrf_name))
        else:
            veth_vrf = constants.OVN_EVPN_VETH_VRF_PREFIX + str(self.vni)
            veth_ovs = constants.OVN_EVPN_VETH_OVS_PREFIX + str(self.vni)
            mock_ensure_veth.assert_called_once_with(veth_vrf, veth_ovs)

            set_master_expected_calls.append(mock.call(veth_vrf, vrf_name))

        mock_set_master.assert_has_calls(set_master_expected_calls)

    def test__ensure_evpn_devices(self):
        self._test__ensure_evpn_devices()

    def test__ensure_evpn_devices_not_vlan(self):
        self._test__ensure_evpn_devices(use_vlan=False)

    @mock.patch.object(linux_net, 'delete_device')
    def test__remove_evpn_devices(self, mock_del_device):
        vrf_name = constants.OVN_EVPN_VRF_PREFIX + str(self.vni)
        bridge_name = constants.OVN_EVPN_BRIDGE_PREFIX + str(self.vni)
        vxlan_name = constants.OVN_EVPN_VXLAN_PREFIX + str(self.vni)
        lo_name = constants.OVN_EVPN_LO_PREFIX + str(self.vni)
        veth_name = constants.OVN_EVPN_VETH_VRF_PREFIX + str(self.vni)
        vlan_name = constants.OVN_EVPN_VLAN_PREFIX + str(self.vni)

        self.evpn_driver._remove_evpn_devices(self.vni)

        expected_calls = [mock.call(lo_name), mock.call(vrf_name),
                          mock.call(bridge_name), mock.call(vxlan_name),
                          mock.call(veth_name), mock.call(vlan_name)]
        mock_del_device.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'add_unreachable_route')
    @mock.patch.object(linux_net, 'add_ndp_proxy')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(ovs, 'add_device_to_ovs_bridge')
    def _test__connect_evpn_to_ovn(
            self, mock_add_ovs_bridge, mock_add_route, mock_ip_version,
            mock_add_ndp_proxy, mock_add_unreachable_route, use_vlan=True):
        mock_ip_version.side_effect = (constants.IP_VERSION_4,
                                       constants.IP_VERSION_6)
        vrf = 'fake-vrf'
        veth_vrf = 'fake-veth-vrf'
        veth_ovs = 'fake-veth-ovs'
        ips = [self.ipv4, self.ipv6]
        dp_bridge = 'datapath-bridge'
        vlan = 'fake-vlan'
        vlan_tag = self.vlan_tag if use_vlan else None

        self.evpn_driver._connect_evpn_to_ovn(
            vrf, veth_vrf, veth_ovs, ips, dp_bridge, self.vni,
            vlan, vlan_tag)

        mock_add_unreachable_route.assert_called_once_with(vrf)

        if not use_vlan:
            mock_add_ndp_proxy.assert_called_once_with(self.ipv6, dp_bridge)
            mock_add_ovs_bridge.assert_called_once_with(veth_ovs, dp_bridge)
            add_route_expected_calls = [
                mock.call(mock.ANY, self.ipv4, self.vni, veth_vrf),
                mock.call(mock.ANY, self.ipv6, self.vni, veth_vrf)]
        else:
            mock_add_ndp_proxy.assert_called_once_with(self.ipv6, vlan)
            add_route_expected_calls = [
                mock.call(mock.ANY, self.ipv4, self.vni, vlan),
                mock.call(mock.ANY, self.ipv6, self.vni, vlan)]

        mock_add_route.assert_has_calls(add_route_expected_calls)

    def test__connect_evpn_to_ovn(self):
        self._test__connect_evpn_to_ovn()

    def test__connect_evpn_to_ovn_not_vlan(self):
        self._test__connect_evpn_to_ovn(use_vlan=False)

    @mock.patch.object(linux_net, 'del_ndp_proxy')
    @mock.patch.object(linux_net, 'delete_routes_from_table')
    @mock.patch.object(ovs, 'del_device_from_ovs_bridge')
    @mock.patch.object(linux_net, 'get_ip_version')
    def _test_disconnect_evpn_from_ovn(
            self, mock_ip_version, mock_del_device, mock_delete_routes,
            mock_del_ndp, use_vlan=True, clean_ndp=True):
        mock_ip_version.side_effect = (constants.IP_VERSION_4,
                                       constants.IP_VERSION_6)
        dp_bridge = 'datapath-bridge'
        ips = [self.ipv4, self.ipv6]
        vlan_tag = self.vlan_tag if use_vlan else None

        self.evpn_driver._disconnect_evpn_from_ovn(
            self.vni, dp_bridge, ips, vlan_tag=vlan_tag,
            cleanup_ndp_proxy=clean_ndp)

        # Assertions
        device = constants.OVN_EVPN_VETH_OVS_PREFIX + str(self.vni)
        if use_vlan:
            device = constants.OVN_EVPN_VLAN_PREFIX + str(self.vni)

        mock_delete_routes.assert_called_once_with(self.vni)

        mock_del_device.assert_called_once_with(device, dp_bridge)
        if clean_ndp:
            mock_del_ndp.assert_called_once_with(self.ipv6, dp_bridge)
        else:
            mock_del_ndp.assert_not_called()

    def test_disconnect_evpn_from_ovn(self):
        self._test_disconnect_evpn_from_ovn()

    def test_disconnect_evpn_from_ovn_dont_clean_ndp(self):
        self._test_disconnect_evpn_from_ovn(clean_ndp=False)

    def test_disconnect_evpn_from_ovn_not_vlan(self):
        self._test_disconnect_evpn_from_ovn(use_vlan=False)

    @mock.patch.object(ovs, 'del_device_from_ovs_bridge')
    @mock.patch.object(linux_net, 'delete_device')
    @mock.patch.object(linux_net, 'get_interfaces')
    def test__remove_extra_vrfs(
            self, mock_get_ifaces, mock_del_device, mock_del_device_bridge):
        # NOTE(lucasagomes): Remove cr_lrp1 to simplify the test
        self.evpn_driver.ovn_local_cr_lrps.pop(self.cr_lrp1, None)
        mock_get_ifaces.return_value = [
            '%siface' % type_ for type_ in (
                constants.OVN_EVPN_VRF_PREFIX,
                constants.OVN_EVPN_LO_PREFIX,
                constants.OVN_EVPN_BRIDGE_PREFIX,
                constants.OVN_EVPN_VXLAN_PREFIX,
                constants.OVN_EVPN_VETH_VRF_PREFIX,
                constants.OVN_EVPN_VLAN_PREFIX)]

        self.evpn_driver._remove_extra_vrfs()

        # Assertions
        expected_calls = [mock.call('vrf-iface'),
                          mock.call('lo-iface'),
                          mock.call('br-iface'),
                          mock.call('vxlan-iface'),
                          mock.call('veth-vrf-iface')]
        mock_del_device.assert_has_calls(expected_calls)

        expected_calls = [mock.call('veth-vrf-iface'),
                          mock.call('vlan-iface')]
        mock_del_device_bridge.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'get_interface_index')
    @mock.patch.object(linux_net, 'delete_ip_routes')
    @mock.patch.object(linux_net, 'get_routes_on_tables')
    def test__remove_extra_routes(
            self, mock_get_routes, mock_del_ip_routes, mock_index):
        mock_index.return_value = 'fake-oif'
        mock_table_ids = mock.patch.object(
            self.evpn_driver, '_get_table_ids').start()
        mock_table_ids.return_value = ['fake-table-id']
        route_to_del = {
            'oif': 'fake-oif0',
            'gateway': 'fake-gateway0',
            'dst': 'fake-dst0',
            'dst_len': 'fake-dst-len0',
            'table': 'fake-table0'}
        mock_get_routes.return_value = [
            self.evpn_driver._ovn_routing_tables_routes[
                'fake-vlan'][0]['route'],
            route_to_del]

        self.evpn_driver._remove_extra_routes()

        # Assert the route meant to be deleted was deleted
        mock_del_ip_routes.assert_called_once_with([route_to_del])

    @mock.patch.object(ovs, 'get_flow_info')
    @mock.patch.object(ovs, 'get_bridge_flows')
    @mock.patch.object(ovs, 'del_flow')
    def test_remove_extra_ovs_flows_mac(
            self, mock_del_flow, mock_get_flows, mock_flow_info):
        mock_flow_info.return_value = {'mac': 'aa:aa:aa:aa:aa:aa'}
        mock_get_flows.return_value = ['fake-flow0', 'fake-flow1']

        self.evpn_driver._remove_extra_ovs_flows()

        expected_calls = [mock.call('fake-flow0', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE),
                          mock.call('fake-flow1', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE)]
        mock_del_flow.assert_has_calls(expected_calls)

    @mock.patch.object(ovs, 'get_flow_info')
    @mock.patch.object(ovs, 'get_bridge_flows')
    @mock.patch.object(ovs, 'del_flow')
    def test_remove_extra_ovs_flows_port(
            self, mock_del_flow, mock_get_flows, mock_flow_info):
        mock_flow_info.return_value = {
            'mac': self.mac,
            'port': 'fake-port',
        }
        mock_get_flows.return_value = ['fake-flow0', 'fake-flow1']

        self.evpn_driver._remove_extra_ovs_flows()

        expected_calls = [mock.call('fake-flow0', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE),
                          mock.call('fake-flow1', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE)]
        mock_del_flow.assert_has_calls(expected_calls)

    @mock.patch.object(ovs, 'get_device_port_at_ovs')
    @mock.patch.object(ovs, 'get_flow_info')
    @mock.patch.object(ovs, 'get_bridge_flows')
    @mock.patch.object(ovs, 'del_flow')
    def test_remove_extra_ovs_flows_port_nw_src(
            self, mock_del_flow, mock_get_flows, mock_flow_info,
            mock_get_port_ovs):
        mock_get_port_ovs.return_value = 'fake-ovs-port'
        mock_flow_info.return_value = {
            'mac': self.mac,
            'port': 'fake-port',
            'nw_src': '10.10.1.88/32',
        }
        mock_get_flows.return_value = ['fake-flow0', 'fake-flow1']

        self.evpn_driver._remove_extra_ovs_flows()

        expected_calls = [mock.call('fake-flow0', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE),
                          mock.call('fake-flow1', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE)]
        mock_del_flow.assert_has_calls(expected_calls)

    @mock.patch.object(ovs, 'get_flow_info')
    @mock.patch.object(ovs, 'get_bridge_flows')
    @mock.patch.object(ovs, 'del_flow')
    def test_remove_extra_ovs_flows(
            self, mock_del_flow, mock_get_flows, mock_flow_info):
        mock_flow_info.return_value = {}
        mock_get_flows.return_value = ['fake-flow0', 'fake-flow1']

        self.evpn_driver._remove_extra_ovs_flows()

        expected_calls = [mock.call('fake-flow0', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE),
                          mock.call('fake-flow1', self.bridge,
                                    constants.OVS_VRF_RULE_COOKIE)]
        mock_del_flow.assert_has_calls(expected_calls)

    @mock.patch.object(linux_net, 'del_ips_from_dev')
    @mock.patch.object(linux_net, 'get_exposed_ips')
    def test__remove_extra_exposed_ips(self, mock_get_ips, mock_del_ips):
        self.evpn_driver._ovn_exposed_evpn_ips = {
            'fake-lo': [self.ipv4, self.ipv6]}
        another_ip = '10.10.1.76'
        mock_get_ips.return_value = [another_ip]

        self.evpn_driver._remove_extra_exposed_ips()

        mock_del_ips.assert_called_once_with('fake-lo', [another_ip])

    def test__get_table_ids(self):
        ret = self.evpn_driver._get_table_ids()
        self.assertEqual([self.vni, self.vni1], ret)

    def test_get_cr_lrp_mac_mapping(self):
        ret = self.evpn_driver._get_cr_lrp_mac_mapping()
        expected_ret = {
            self.mac: {
                'veth_ovs': 'fake-veth-ovs',
                'veth_vrf': 'fake-veth-vrf',
                'vlan': 'fake-vlan'},
            self.mac1: {
                'veth_ovs': 'fake-veth-ovs1',
                'veth_vrf': 'fake-veth-vrf1',
                'vlan': 'fake-vlan1'}
        }
        self.assertEqual(expected_ret, ret)



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import evpn as evpn_utils
from ovn_bgp_agent.drivers.openstack.utils import ovn as ovn_utils
from ovn_bgp_agent.drivers.openstack.utils import ovs as ovs_utils
from ovn_bgp_agent.drivers.openstack.utils import wire
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils as test_utils
from ovn_bgp_agent.utils import linux_net

CONF = cfg.CONF


class TestWire(test_base.TestCase):

    def setUp(self):
        super(TestWire, self).setUp()
        self.nb_idl = ovn_utils.OvsdbNbOvnIdl(mock.Mock())
        self.sb_idl = ovn_utils.OvsdbSbOvnIdl(mock.Mock())
        self.ovs_idl = mock.Mock()

        # Helper variables that are used across multiple methods
        self.bridge_mappings = ['datacentre:br-ex']

        self.ovs_idl.get_ovn_bridge_mappings.return_value = (
            self.bridge_mappings)

        # Monkey-patch parent class methods
        self.nb_idl.ls_add = mock.Mock()
        self.nb_idl.lr_add = mock.Mock()
        self.nb_idl.lrp_add = mock.Mock()
        self.nb_idl.lrp_set_gateway_chassis = mock.Mock()
        self.nb_idl.lrp_add_networks = mock.Mock()
        self.nb_idl.lr_route_add = mock.Mock()
        self.nb_idl.lr_policy_add = mock.Mock()

    @mock.patch.object(wire, '_ensure_base_wiring_config_underlay')
    def test_ensure_base_wiring_config(self, mock_underlay):
        wire.ensure_base_wiring_config(self.sb_idl, self.ovs_idl,
                                       routing_tables={})
        mock_underlay.assert_called_once_with(self.sb_idl, self.ovs_idl, {})

    @mock.patch.object(wire, '_ensure_base_wiring_config_ovn')
    def test_ensure_base_wiring_config_ovn(self, mock_ovn):
        CONF.set_override('exposing_method', 'ovn')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        wire.ensure_base_wiring_config(self.sb_idl, self.ovs_idl,
                                       ovn_idl=self.nb_idl)
        mock_ovn.assert_called_once_with(self.ovs_idl, self.nb_idl)

    @mock.patch.object(wire, '_ensure_base_wiring_config_evpn')
    def test_ensure_base_wiring_config_evpn(self, mock_evpn):
        CONF.set_override('exposing_method', 'vrf')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        wire.ensure_base_wiring_config(self.nb_idl, self.ovs_idl,
                                       ovn_idl=self.nb_idl)
        mock_evpn.assert_called_once_with(self.nb_idl, self.ovs_idl)

    @mock.patch.object(wire, '_ensure_base_wiring_config_underlay')
    @mock.patch.object(wire, '_ensure_base_wiring_config_ovn')
    def test_ensure_base_wiring_config_not_implemented(self, mock_ovn,
                                                       mock_underlay):
        CONF.set_override('exposing_method', 'dynamic')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        self.assertRaises(agent_exc.UnsupportedWiringConfig,
                          wire.ensure_base_wiring_config,
                          self.sb_idl, self.ovs_idl, ovn_idl=self.nb_idl)
        mock_ovn.assert_not_called()
        mock_underlay.assert_not_called()

    def test__ensure_base_wiring_config_ovn(self):
        pass

    @mock.patch.object(linux_net, 'get_interface_address')
    @mock.patch.object(ovs_utils, 'get_ovs_patch_ports_info')
    def test__ensure_base_wiring_config_evpn(self, m_get_ovs_patch_ports_info,
                                             m_get_interface_address):
        localnet_ports = [test_utils.create_row(
            tag=[4096],
        )]
        ports = localnet_ports + [test_utils.create_row(
            dhcpv4_options=[test_utils.create_row()],
            dhcpv6_options=[],
        )]
        get_localnet_ports_by_network_name = mock.patch.object(
            self.nb_idl, 'get_localnet_ports_by_network_name').start()
        get_localnet_ports_by_network_name.return_value = localnet_ports

        provnets = [test_utils.create_row(
            name='fake-provnet',
            external_ids={
                constants.OVN_EVPN_VNI_EXT_ID_KEY: 100,
            },
            ports=ports,
        )]
        get_bgpvpn_networks_for_ports = mock.patch.object(
            self.nb_idl, 'get_bgpvpn_networks_for_ports').start()
        get_bgpvpn_networks_for_ports.return_value = provnets

        CONF.set_override('evpn_local_ip', '127.0.0.1')
        self.addCleanup(CONF.clear_override, 'evpn_local_ip')

        vlan_dev = mock.MagicMock()

        evpn_bridge = mock.MagicMock()
        evpn_bridge.connect_vlan.return_value = vlan_dev

        evpn_setup = mock.patch.object(evpn_utils, 'setup').start()
        evpn_setup.return_value = evpn_bridge

        wire._ensure_base_wiring_config_evpn(self.nb_idl, self.ovs_idl)

        evpn_setup.assert_called_with(ovs_bridge='br-ex',
                                      vni=100,
                                      evpn_opts={'route_targets': [],
                                                 'route_distinguishers': [],
                                                 'export_targets': [],
                                                 'import_targets': []},
                                      mode=constants.OVN_EVPN_TYPE_L3,
                                      ovs_flows=mock.ANY)

        evpn_bridge.connect_vlan.assert_called_with(ports[0])
        vlan_dev.process_dhcp_opts.assert_called()

    def test__ensure_ovn_router(self):
        wire._ensure_ovn_router(self.nb_idl)
        self.nb_idl.lr_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, may_exist=True)

    @mock.patch.object(wire, '_execute_commands')
    @mock.patch.object(wire, '_ensure_lsp_cmds')
    def test__ensure_ovn_switch(self, m_ensure_lsp, m_cmds):
        ls_name = 'test-ls'
        localnet_port = "{}-localnet".format(ls_name)
        options = {'network_name': ls_name}

        wire._ensure_ovn_switch(self.nb_idl, ls_name)
        self.nb_idl.ls_add.assert_called_once_with(ls_name, may_exist=True)
        m_ensure_lsp.assert_called_once_with(
            self.nb_idl, localnet_port, ls_name, 'localnet', 'unknown',
            **options)

    @mock.patch.object(wire, '_execute_commands')
    @mock.patch.object(wire, '_ensure_lsp_cmds')
    def test__ensure_ovn_network_link_internal(self, m_ensure_lsp, m_cmds):
        switch_name = 'internal'
        provider_cidrs = ['172.16.0.0/16']
        r_port_name = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
        options = {'router-port': r_port_name, 'arp_proxy': '172.16.0.0/16'}

        wire._ensure_ovn_network_link_internal(
            self.nb_idl, switch_name, provider_cidrs)

        self.nb_idl.lrp_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, r_port_name,
            constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC, provider_cidrs, peer=[],
            may_exist=True)
        m_ensure_lsp.assert_called_once_with(
            self.nb_idl, mock.ANY, switch_name, 'router', 'router', **options)
        self.nb_idl.lrp_set_gateway_chassis.assert_called_once_with(
            r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1)

    @mock.patch.object(wire, '_execute_commands')
    @mock.patch.object(wire, '_ensure_lsp_cmds')
    def test__ensure_ovn_network_link_internal_runtime_error(
            self, m_ensure_lsp, m_cmds):
        switch_name = 'internal'
        provider_cidrs = ['172.16.0.0/16']
        r_port_name = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)
        options = {'router-port': r_port_name, 'arp_proxy': '172.16.0.0/16'}
        self.nb_idl.lrp_add.side_effect = RuntimeError(
            'with different networks')

        wire._ensure_ovn_network_link_internal(
            self.nb_idl, switch_name, provider_cidrs)

        self.nb_idl.lrp_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, r_port_name,
            constants.OVN_CLUSTER_ROUTER_INTERNAL_MAC, provider_cidrs, peer=[],
            may_exist=True)
        self.nb_idl.lrp_add_networks.assert_called_once_with(
            r_port_name, provider_cidrs, may_exist=True)
        m_ensure_lsp.assert_called_once_with(
            self.nb_idl, mock.ANY, switch_name, 'router', 'router', **options)
        self.nb_idl.lrp_set_gateway_chassis.assert_called_once_with(
            r_port_name, CONF.local_ovn_cluster.bgp_chassis_id, 1)

    @mock.patch.object(wire, '_execute_commands')
    @mock.patch.object(wire, '_ensure_lsp_cmds')
    def test__ensure_ovn_network_link_external(
            self, m_ensure_lsp, m_cmds):
        switch_name = 'external'
        ip = '1.1.1.2'
        mac = 'fake-map'
        r_port_name = "{}-{}".format(constants.OVN_CLUSTER_ROUTER, switch_name)
        options = {'router-port': r_port_name}

        wire._ensure_ovn_network_link_external(
            self.nb_idl, switch_name, ip, mac)

        self.nb_idl.lrp_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, r_port_name,
            mac, [ip], peer=[], may_exist=True)
        m_ensure_lsp.assert_called_once_with(
            self.nb_idl, mock.ANY, switch_name, 'router', 'router', **options)

    def _ensure_ovn_policies(self, next_hops):
        if len(next_hops) > 1:
            columns = {'nexthops': next_hops}
        elif len(next_hops) == 1:
            columns = {'nexthop': next_hops[0]}

        wire._ensure_ovn_policies(self.nb_idl, next_hops)

        self.nb_idl.lr_policy_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, 10, mock.ANY, 'reroute',
            may_exist=True, **columns)

    def test__ensure_ovn_policies_dual_nexthop(self):
        next_hops = ['1.1.1.1', '2.2.2.2']
        self._ensure_ovn_policies(next_hops)

    def test__ensure_ovn_policies_single_nexthop(self):
        next_hops = ['1.1.1.1']
        self._ensure_ovn_policies(next_hops)

    @mock.patch.object(wire, '_execute_commands')
    def test_ensure_ovn_routes(self, m_cmds):
        peer_ips = ['1.1.1.1']
        bfds = [test_utils.create_row(uuid='bfd')]
        wire._ensure_ovn_routes(self.nb_idl, peer_ips, bfds)
        self.nb_idl.lr_route_add.assert_called_once_with(
            constants.OVN_CLUSTER_ROUTER, '0.0.0.0/0', peer_ips[0],
            ecmp=True, may_exist=True, bfd=bfds[0].uuid)

    @mock.patch.object(ovs_utils, 'ensure_flow')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(ovs_utils, 'get_ovs_ports_info')
    @mock.patch.object(ovs_utils, 'get_ovs_patch_ports_info')
    def test__ensure_ingress_flows(self, m_ovs_patch, m_ovs_get, m_ip_version,
                                   m_ovn_flow):
        CONF.set_override('external_nics', ['eth1'], group='local_ovn_cluster')
        self.addCleanup(CONF.clear_override, 'external_nics',
                        group='local_ovn_cluster')
        bridge = 'br-ex'
        mac = 'fake-mac'
        switch_name = 'test-ls'
        provider_cidrs = ['172.16.0.0/16']
        patch_port_prefix = 'patch-{}-'.format(switch_name)
        m_ovs_patch.return_value = ['fake-patch']
        m_ovs_get.return_value = ['eth1']
        wire._ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs)
        m_ovs_patch.assert_called_once_with(bridge, prefix=patch_port_prefix)
        m_ovs_get.assert_called_once_with(bridge)
        m_ip_version.assert_called_once_with(provider_cidrs[0])
        m_ovn_flow.assert_called_once_with(bridge, mock.ANY)

    @mock.patch.object(ovs_utils, 'get_ovs_patch_ports_info')
    def test__ensure_ingress_flows_no_network(self, m_ovs):
        bridge = 'br-ex'
        mac = 'fake-mac'
        switch_name = 'test-ls'
        provider_cidrs = []
        wire._ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs)
        m_ovs.assert_not_called()

    @mock.patch.object(ovs_utils, 'get_ovs_ports_info')
    @mock.patch.object(ovs_utils, 'get_ovs_patch_ports_info')
    def test__ensure_ingress_flows_no_patch_port(self, m_ovs_patch, m_ovs_get):
        bridge = 'br-ex'
        mac = 'fake-mac'
        switch_name = 'test-ls'
        provider_cidrs = ['172.16.0.0/16']
        patch_port_prefix = 'patch-{}-'.format(switch_name)
        m_ovs_patch.return_value = []
        wire._ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs)
        m_ovs_patch.assert_called_once_with(bridge, prefix=patch_port_prefix)
        m_ovs_get.assert_not_called()

    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(ovs_utils, 'get_ovs_ports_info')
    @mock.patch.object(ovs_utils, 'get_ovs_patch_ports_info')
    def test__ensure_ingress_flows_no_external_nic(self, m_ovs_patch,
                                                   m_ovs_get, m_ip_version):
        bridge = 'br-ex'
        mac = 'fake-mac'
        switch_name = 'test-ls'
        provider_cidrs = ['172.16.0.0/16']
        patch_port_prefix = 'patch-{}-'.format(switch_name)
        m_ovs_patch.return_value = ['fake-patch']
        m_ovs_get.return_value = ['eth1']
        wire._ensure_ingress_flows(bridge, mac, switch_name, provider_cidrs)
        m_ovs_patch.assert_called_once_with(bridge, prefix=patch_port_prefix)
        m_ovs_get.assert_called_once_with(bridge)
        m_ip_version.assert_not_called()

    @mock.patch.object(wire, '_cleanup_wiring_underlay')
    def test_cleanup_wiring_underlay(self, mock_underlay):
        ovs_flows = {}
        exposed_ips = {}
        routing_tables = {}
        routing_tables_routes = {}
        wire.cleanup_wiring(self.sb_idl, self.bridge_mappings, ovs_flows,
                            exposed_ips, routing_tables, routing_tables_routes)
        mock_underlay.assert_called_once_with(
            self.sb_idl, self.bridge_mappings, ovs_flows, exposed_ips,
            routing_tables, routing_tables_routes)

    @mock.patch.object(linux_net, 'get_exposed_ips')
    @mock.patch.object(linux_net, 'delete_exposed_ips')
    def test_cleanup_wiring_ovn(self, _delete_exposed_ips, _get_exposed_ips):
        CONF.set_override('exposing_method', 'ovn')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        # Both IPs are present on the NIC but .42 is not expected
        # to be exposed.
        _get_exposed_ips.return_value = ['192.0.2.42', '192.0.2.24']

        ovs_flows = {}
        exposed_ips = {
            "fakels": {"192.0.2.24": {'bridge_device': "br-example"}}}
        routing_tables = {}
        routing_tables_routes = {}
        wire.cleanup_wiring(self.sb_idl, self.bridge_mappings, ovs_flows,
                            exposed_ips, routing_tables,
                            routing_tables_routes)
        # Make sure we only delete the IP that isn't supposed to be exposed.
        _delete_exposed_ips.assert_called_once_with({"192.0.2.42"},
                                                    CONF.bgp_nic)

    def test_cleanup_wiring_evpn(self):
        CONF.set_override('exposing_method', 'vrf')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        vlan_dev = mock.MagicMock()
        evpn_bridge = mock.MagicMock()
        evpn_bridge.get_vlan.return_value = vlan_dev
        ovs_flows = {
            'foo': {
                'evpn': {
                    '4096': evpn_bridge,
                }
            }
        }
        exposed_ips = {}
        routing_tables = {}
        routing_tables_routes = {}
        ret = wire.cleanup_wiring(self.sb_idl, self.bridge_mappings, ovs_flows,
                                  exposed_ips, routing_tables,
                                  routing_tables_routes)

        evpn_bridge.get_vlan.assert_called_with('4096')
        vlan_dev.cleanup_excessive_routes.assert_called()
        self.assertTrue(ret)

    @mock.patch.object(wire, '_cleanup_wiring_underlay')
    def test_cleanup_wiring_not_implemeneted(self, mock_underlay):
        CONF.set_override('exposing_method', 'dynamic')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        ovs_flows = {}
        exposed_ips = {}
        routing_tables = {}
        routing_tables_routes = {}
        wire.cleanup_wiring(self.sb_idl, self.bridge_mappings, ovs_flows,
                            exposed_ips, routing_tables, routing_tables_routes)

        mock_underlay.assert_not_called()

    @mock.patch.object(wire, '_wire_provider_port_underlay')
    def test_wire_provider_port_underlay(self, mock_underlay):
        routing_tables_routes = {}
        ovs_flows = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        localnet = 'fake-localnet'
        routing_table = 5
        proxy_cidrs = []

        wire.wire_provider_port(routing_tables_routes, ovs_flows, port_ips,
                                bridge_device, bridge_vlan, localnet,
                                routing_table, proxy_cidrs)
        mock_underlay.assert_called_once_with(
            routing_tables_routes, ovs_flows, port_ips, bridge_device,
            bridge_vlan, localnet, routing_table, proxy_cidrs, lladdr=None)

    @mock.patch.object(wire, '_wire_provider_port_ovn')
    def test_wire_provider_port_ovn(self, mock_ovn):
        CONF.set_override('exposing_method', 'ovn')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        ovs_flows = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        localnet = 'fake-localnet'
        routing_table = 5
        proxy_cidrs = []
        mac = 'fake-mac'

        wire.wire_provider_port(routing_tables_routes, ovs_flows, port_ips,
                                bridge_device, bridge_vlan, localnet,
                                routing_table, proxy_cidrs, mac=mac,
                                ovn_idl=self.nb_idl)
        mock_ovn.assert_called_once_with(self.nb_idl, port_ips, mac)

    def test_wire_provider_port_evpn(self):
        CONF.set_override('exposing_method', 'vrf')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        ovs_flows = {}
        port_ips = ['10.10.10.1']
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        localnet = 'fake-localnet'
        routing_table = 5
        proxy_cidrs = []
        mac = 'fake-mac'

        vlan_dev = mock.MagicMock()

        evpn_bridge = mock.MagicMock()
        evpn_bridge.get_vlan.return_value = vlan_dev

        evpn_lookup = mock.patch.object(evpn_utils, 'lookup').start()
        evpn_lookup.return_value = evpn_bridge

        ret = wire.wire_provider_port(routing_tables_routes, ovs_flows,
                                      port_ips, bridge_device, bridge_vlan,
                                      localnet, routing_table, proxy_cidrs,
                                      mac=mac, ovn_idl=self.nb_idl)
        self.assertTrue(ret)

        evpn_lookup.assert_called_once_with(bridge_device, bridge_vlan)
        evpn_bridge.get_vlan.assert_called_once_with(bridge_vlan)
        vlan_dev.add_route.assert_called_with(routing_tables_routes,
                                              port_ips[0], mac, via=None)

    def test_wire_provider_port_evpn_unconfigured(self):
        CONF.set_override('exposing_method', 'vrf')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        ovs_flows = {}
        port_ips = ['10.10.10.1']
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        localnet = 'fake-localnet'
        routing_table = 5
        proxy_cidrs = []
        mac = 'fake-mac'

        evpn_lookup = mock.patch.object(evpn_utils, 'lookup').start()
        evpn_lookup.side_effect = KeyError

        ret = wire.wire_provider_port(routing_tables_routes, ovs_flows,
                                      port_ips, bridge_device, bridge_vlan,
                                      localnet, routing_table, proxy_cidrs,
                                      mac=mac, ovn_idl=self.nb_idl)
        self.assertIsNone(ret)

    @mock.patch.object(wire, '_wire_provider_port_underlay')
    @mock.patch.object(wire, '_wire_provider_port_ovn')
    @mock.patch.object(wire, '_wire_provider_port_evpn')
    def test_wire_provider_port_not_implemented(self, mock_evpn, mock_ovn,
                                                mock_underlay):
        CONF.set_override('exposing_method', 'dynamic')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        ovs_flows = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        localnet = 'fake-localnet'
        routing_table = 5
        proxy_cidrs = []

        wire.wire_provider_port(routing_tables_routes, ovs_flows, port_ips,
                                bridge_device, bridge_vlan, localnet,
                                routing_table, proxy_cidrs)

        mock_evpn.assert_not_called()
        mock_ovn.assert_not_called()
        mock_underlay.assert_not_called()

    @mock.patch.object(wire, '_execute_commands')
    def test__wire_provider_port_ovn(self, m_cmds):
        port_ips = ['1.1.1.1', '2.2.2.2']
        mac = 'fake-mac'
        port = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)

        wire._wire_provider_port_ovn(self.nb_idl, port_ips, mac)

        cmds = [
            ovn_utils.StaticMACBindingAddCommand(
                self.nb_idl, port, port_ips[0], mac, True, may_exist=True),
            ovn_utils.StaticMACBindingAddCommand(
                self.nb_idl, port, port_ips[1], mac, True, may_exist=True)]

        # FIXME(ltomasbo): The standard assert called ones is not working
        # with the object
        # m_cmds.assert_called_once_with(self.nb_idl, cmds)
        # so we are checking this by comparing the object dict instead
        self.assertEqual(
            m_cmds.call_args_list[0][0][1][0].__dict__,
            cmds[0].__dict__
        )

    @mock.patch.object(wire, '_execute_commands')
    def test__wire_provider_port_ovn_no_action(self, m_cmds):
        port_ips = []
        mac = 'fake-mac'
        wire._wire_provider_port_ovn(self.nb_idl, port_ips, mac)
        m_cmds.assert_not_called()

    @mock.patch.object(wire, '_unwire_provider_port_underlay')
    def test_unwire_provider_port_underlay(self, mock_underlay):
        routing_tables_routes = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_table = 5
        proxy_cidrs = []

        wire.unwire_provider_port(routing_tables_routes, port_ips,
                                  bridge_device, bridge_vlan, routing_table,
                                  proxy_cidrs)
        mock_underlay.assert_called_once_with(
            routing_tables_routes, port_ips, bridge_device, bridge_vlan,
            routing_table, proxy_cidrs, lladdr=None)

    @mock.patch.object(wire, '_unwire_provider_port_ovn')
    def test_unwire_provider_port_ovn(self, mock_ovn):
        CONF.set_override('exposing_method', 'ovn')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_table = 5
        proxy_cidrs = []

        wire.unwire_provider_port(routing_tables_routes, port_ips,
                                  bridge_device, bridge_vlan, routing_table,
                                  proxy_cidrs, ovn_idl=self.nb_idl)
        mock_ovn.assert_called_once_with(self.nb_idl, port_ips)

    @mock.patch.object(wire, '_unwire_provider_port_evpn')
    def test_unwire_provider_port_evpn(self, mock_evpn):
        CONF.set_override('exposing_method', 'vrf')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_table = 5
        proxy_cidrs = []

        wire.unwire_provider_port(routing_tables_routes, port_ips,
                                  bridge_device, bridge_vlan, routing_table,
                                  proxy_cidrs, mac='boo')
        mock_evpn.assert_called_once_with(routing_tables_routes, port_ips,
                                          bridge_device, bridge_vlan, 'boo')

    @mock.patch.object(wire, '_unwire_provider_port_underlay')
    @mock.patch.object(wire, '_unwire_provider_port_ovn')
    @mock.patch.object(wire, '_unwire_provider_port_evpn')
    def test_unwire_provider_port_not_implemented(self, mock_evpn, mock_ovn,
                                                  mock_underlay):
        CONF.set_override('exposing_method', 'dynamic')
        self.addCleanup(CONF.clear_override, 'exposing_method')

        routing_tables_routes = {}
        port_ips = []
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_table = 5
        proxy_cidrs = []

        wire.unwire_provider_port(routing_tables_routes, port_ips,
                                  bridge_device, bridge_vlan, routing_table,
                                  proxy_cidrs)
        mock_evpn.assert_not_called()
        mock_ovn.assert_not_called()
        mock_underlay.assert_not_called()

    def test__unwire_provider_port_evpn(self):
        routing_tables_routes = {}
        port_ips = ['10.10.10.1']
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        lladdr = 'boo'

        vlan_dev = mock.MagicMock()

        evpn_bridge = mock.MagicMock()
        evpn_bridge.get_vlan.return_value = vlan_dev

        evpn_lookup = mock.patch.object(evpn_utils, 'lookup').start()
        evpn_lookup.return_value = evpn_bridge

        ret = wire._unwire_provider_port_evpn(routing_tables_routes, port_ips,
                                              bridge_device, bridge_vlan,
                                              lladdr)
        self.assertTrue(ret)

        vlan_dev.del_route.assert_called_with(routing_tables_routes,
                                              port_ips[0], lladdr)

    def test__unwire_provider_port_evpn_unconfigured(self):
        routing_tables_routes = {}
        port_ips = ['10.10.10.1']
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        lladdr = 'boo'

        evpn_lookup = mock.patch.object(evpn_utils, 'lookup').start()
        evpn_lookup.side_effect = KeyError

        ret = wire._unwire_provider_port_evpn(routing_tables_routes, port_ips,
                                              bridge_device, bridge_vlan,
                                              lladdr)
        self.assertIsNone(ret)

    @mock.patch.object(wire, '_execute_commands')
    def test__unwire_provider_port_ovn(self, m_cmds):
        port_ips = ['1.1.1.1']
        port = "{}-openstack".format(constants.OVN_CLUSTER_ROUTER)

        wire._unwire_provider_port_ovn(self.nb_idl, port_ips)

        cmds = [ovn_utils.StaticMACBindingDelCommand(
                self.nb_idl, port, port_ips[0], if_exists=True)]

        # FIXME(ltomasbo): The standard assert called ones is not working
        # with the object
        # m_cmds.assert_called_once_with(self.nb_idl, cmds)
        # so we are checking this by comparing the object dict instead
        self.assertEqual(
            m_cmds.call_args_list[0][0][1][0].__dict__,
            cmds[0].__dict__
        )

    @mock.patch.object(wire, '_execute_commands')
    def test__unwire_provider_port_ovn_no_action(self, m_cmds):
        port_ips = []
        wire._unwire_provider_port_ovn(self.nb_idl, port_ips)
        m_cmds.assert_not_called()

    @mock.patch.object(wire, '_wire_lrp_port_underlay')
    def test_wire_lrp_port_underlay(self, mock_underlay):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        wire.wire_lrp_port(routing_tables_routes, ip, bridge_device,
                           bridge_vlan, routing_tables, cr_lrp_ips)
        mock_underlay.assert_called_once_with(
            routing_tables_routes, ip, bridge_device, bridge_vlan,
            routing_tables, cr_lrp_ips)

    @mock.patch.object(wire, '_unwire_lrp_port_underlay')
    def test_unwire_lrp_port_underlay(self, mock_underlay):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        wire.unwire_lrp_port(routing_tables_routes, ip, bridge_device,
                             bridge_vlan, routing_tables, cr_lrp_ips)
        mock_underlay.assert_called_once_with(
            routing_tables_routes, ip, bridge_device, bridge_vlan,
            routing_tables, cr_lrp_ips)

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__wire_lrp_port_underlay(self, m_ip_rule, m_ip_version,
                                     m_ip_route):
        routing_tables_routes = {}
        ip = '10.0.0.1/24'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        ret = wire._wire_lrp_port_underlay(routing_tables_routes, ip,
                                           bridge_device, bridge_vlan,
                                           routing_tables, cr_lrp_ips)
        self.assertTrue(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        m_ip_route.assert_called_once_with(
            routing_tables_routes, '10.0.0.1', 5, 'fake-bridge',
            vlan='101', mask='24', via='fake-crlrp-ip')

    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__wire_lrp_port_underlay_no_bridge(self, m_ip_rule):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = None
        bridge_vlan = None
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        ret = wire._wire_lrp_port_underlay(routing_tables_routes, ip,
                                           bridge_device, bridge_vlan,
                                           routing_tables, cr_lrp_ips)

        self.assertFalse(ret)
        m_ip_rule.assert_not_called()

    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__wire_lrp_port_underlay_invalid_ip(self, m_ip_rule, m_ip_version):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']
        m_ip_rule.side_effect = agent_exc.InvalidPortIP(ip=ip)

        ret = wire._wire_lrp_port_underlay(routing_tables_routes, ip,
                                           bridge_device, bridge_vlan,
                                           routing_tables, cr_lrp_ips)

        self.assertFalse(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        m_ip_version.assert_not_called()

    @mock.patch.object(linux_net, 'add_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'add_ip_rule')
    def test__wire_lrp_port_underlay_advertisement_subnet(
        self, m_ip_rule, m_ip_version, m_ip_route):
        CONF.set_override(
            'advertisement_method_tenant_networks',
            constants.ADVERTISEMENT_METHOD_SUBNET)
        self.addCleanup(
            CONF.clear_override,
            'advertisement_method_tenant_networks')
        routing_tables_routes = {}
        ip = '10.0.0.1/24'
        bridge_device = 'fake-bridge'
        bridge_vlan = None
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']
        ret = wire._wire_lrp_port_underlay(
            routing_tables_routes, ip, bridge_device,
            bridge_vlan, routing_tables, cr_lrp_ips)
        self.assertTrue(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        expected_ip_route_calls = [
            mock.call(
                routing_tables_routes, ip.split('/')[0],
                routing_tables[bridge_device], bridge_device,
                vlan=bridge_vlan, mask=ip.split('/')[1], via=cr_lrp_ips[0]),
            mock.call(
                routing_tables_routes, ip.split('/')[0],
                CONF.bgp_vrf_table_id, CONF.bgp_nic,
                mask=ip.split('/')[1], via=cr_lrp_ips[0])]
        m_ip_route.assert_has_calls(expected_ip_route_calls)

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__unwire_lrp_port_underlay(self, m_ip_rule, m_ip_version,
                                       m_ip_route):
        routing_tables_routes = {}
        ip = '10.0.0.1/24'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        ret = wire._unwire_lrp_port_underlay(routing_tables_routes, ip,
                                             bridge_device, bridge_vlan,
                                             routing_tables, cr_lrp_ips)
        self.assertTrue(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        m_ip_route.assert_called_once_with(
            routing_tables_routes, '10.0.0.1', 5, 'fake-bridge',
            vlan='101', mask='24', via='fake-crlrp-ip')

    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__unwire_lrp_port_underlay_no_bridge(self, m_ip_rule):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = None
        bridge_vlan = None
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']

        ret = wire._unwire_lrp_port_underlay(routing_tables_routes, ip,
                                             bridge_device, bridge_vlan,
                                             routing_tables, cr_lrp_ips)

        self.assertFalse(ret)
        m_ip_rule.assert_not_called()

    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__unwire_lrp_port_underlay_invalid_ip(self, m_ip_rule,
                                                  m_ip_version):
        routing_tables_routes = {}
        ip = 'fake-ip'
        bridge_device = 'fake-bridge'
        bridge_vlan = '101'
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']
        m_ip_rule.side_effect = agent_exc.InvalidPortIP(ip=ip)

        ret = wire._unwire_lrp_port_underlay(routing_tables_routes, ip,
                                             bridge_device, bridge_vlan,
                                             routing_tables, cr_lrp_ips)

        self.assertFalse(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        m_ip_version.assert_not_called()

    @mock.patch.object(linux_net, 'del_ip_route')
    @mock.patch.object(linux_net, 'get_ip_version')
    @mock.patch.object(linux_net, 'del_ip_rule')
    def test__unwire_lrp_port_underlay_advertisement_subnet(
        self, m_ip_rule, m_ip_version, m_ip_route):
        CONF.set_override(
            'advertisement_method_tenant_networks',
            constants.ADVERTISEMENT_METHOD_SUBNET)
        self.addCleanup(
            CONF.clear_override,
            'advertisement_method_tenant_networks')
        routing_tables_routes = {}
        ip = '10.0.0.1/24'
        bridge_device = 'fake-bridge'
        bridge_vlan = None
        routing_tables = {'fake-bridge': 5}
        cr_lrp_ips = ['fake-crlrp-ip']
        ret = wire._unwire_lrp_port_underlay(
            routing_tables_routes, ip, bridge_device, bridge_vlan,
            routing_tables, cr_lrp_ips)
        self.assertTrue(ret)
        m_ip_rule.assert_called_once_with(ip, 5)
        expected_ip_route_calls = [
            mock.call(
                routing_tables_routes, ip.split('/')[0],
                routing_tables[bridge_device], bridge_device,
                vlan=bridge_vlan, mask=ip.split('/')[1], via=cr_lrp_ips[0]),
            mock.call(
                routing_tables_routes, ip.split('/')[0],
                CONF.bgp_vrf_table_id, CONF.bgp_nic,
                mask=ip.split('/')[1], via=cr_lrp_ips[0])]
        m_ip_route.assert_has_calls(expected_ip_route_calls)



# Copyright 2024 team.blue/nl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from unittest import mock

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import bgp as bgp_utils
from ovn_bgp_agent.tests import base as test_base


CONF = cfg.CONF


class TestEVPN(test_base.TestCase):

    def setUp(self):
        super(TestEVPN, self).setUp()

        self.mock_frr = mock.patch.object(bgp_utils, 'frr').start()
        self.mock_linux_net = mock.patch.object(bgp_utils, 'linux_net').start()

    def _set_exposing_method(self, exposing_method):
        CONF.set_override('exposing_method', exposing_method)
        self.addCleanup(CONF.clear_override, 'exposing_method')

    def _test_announce_ips(self, exposing_method):
        ips = ['10.10.10.1', '10.20.10.1']
        self._set_exposing_method(exposing_method)

        bgp_utils.announce_ips(list(ips))

        if exposing_method in [constants.EXPOSE_METHOD_VRF]:
            self.mock_linux_net.add_ips_to_dev.assert_not_called()
        else:
            self.mock_linux_net.add_ips_to_dev.assert_called_once_with(
                CONF.bgp_nic, ips)

    def test_announce_ips_underlay(self):
        self._test_announce_ips('underlay')

    def test_announce_ips_dynamic(self):
        self._test_announce_ips('dynamic')

    def test_announce_ips_ovn(self):
        self._test_announce_ips('ovn')

    def test_announce_ips_vrf(self):
        self._test_announce_ips('vrf')

    def test_announce_ips_l2vni(self):
        self._test_announce_ips('l2vni')

    def _test_withdraw_ips(self, exposing_method):
        ips = ['10.10.10.1', '10.20.10.1']
        self._set_exposing_method(exposing_method)

        bgp_utils.withdraw_ips(list(ips))

        if exposing_method in [constants.EXPOSE_METHOD_VRF]:
            self.mock_linux_net.del_ips_from_dev.assert_not_called()
        else:
            self.mock_linux_net.del_ips_from_dev.assert_called_once_with(
                CONF.bgp_nic, ips)

    def test_withdraw_ips_underlay(self):
        self._test_withdraw_ips('underlay')

    def test_withdraw_ips_dynamic(self):
        self._test_withdraw_ips('dynamic')

    def test_withdraw_ips_ovn(self):
        self._test_withdraw_ips('ovn')

    def test_withdraw_ips_vrf(self):
        self._test_withdraw_ips('vrf')

    def test_withdraw_ips_l2vni(self):
        self._test_withdraw_ips('l2vni')

    def _test_ensure_base_bgp_configuration(self, exposing_method):
        self._set_exposing_method(exposing_method)

        bgp_utils.ensure_base_bgp_configuration()

        if exposing_method not in [constants.EXPOSE_METHOD_UNDERLAY,
                                   constants.EXPOSE_METHOD_DYNAMIC,
                                   constants.EXPOSE_METHOD_OVN]:
            self.mock_frr.vrf_leak.assert_not_called()
            self.mock_linux_net.ensure_vrf.assert_not_called()
            self.mock_linux_net.ensure_ovn_device.assert_not_called()
        else:
            self.mock_frr.vrf_leak.assert_called_once()
            self.mock_linux_net.ensure_vrf.assert_called_once()
            self.mock_linux_net.ensure_ovn_device.assert_called_once()

    def test_ensure_base_bgp_configuration_underlay(self):
        self._test_ensure_base_bgp_configuration('underlay')

    def test_ensure_base_bgp_configuration_dynamic(self):
        self._test_ensure_base_bgp_configuration('dynamic')

    def test_ensure_base_bgp_configuration_ovn(self):
        self._test_ensure_base_bgp_configuration('ovn')

    def test_ensure_base_bgp_configuration_vrf(self):
        self._test_ensure_base_bgp_configuration('vrf')

    def test_ensure_base_bgp_configuration_l2vni(self):
        self._test_ensure_base_bgp_configuration('l2vni')



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib._i18n import _


class OVNBGPAgentException(Exception):
    """Base OVN BGP Agebt Exception.

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """

    message = _("An unknown exception occurred.")

    def __init__(self, **kwargs):
        super().__init__(self.message % kwargs)
        self.msg = self.message % kwargs

    def __str__(self):
        return self.msg


class ConfOptionRequired(OVNBGPAgentException):
    """Configuration option is required for this driver to function properly

    :param option: The configuration option required
    """

    message = _("Required driver conf option missing: CONF.%(option)s")


class UnsupportedWiringConfig(OVNBGPAgentException):
    """Wiring config called with unknown/unsupported exposing method

    :param method: The (unsupported) exposing method
    """

    message = _("No wiring support for exposing_method %(method)s.")


class InvalidPortIP(OVNBGPAgentException):
    """OVN Port has Invalid IP.

    :param ip: The (wrong) IP of the port
    """

    message = _("OVN port with invalid IP: %(ip)s.")


class PortNotFound(OVNBGPAgentException):
    """OVN Port not found.

    :param port: The port name or UUID.
    """

    message = _("OVN port was not found: %(port)s.")


class DatapathNotFound(OVNBGPAgentException):
    """Datapath not found

    :param datapath: The datapath UUID
    """

    message = _("Datapath was not found: %(datapath)s.")


class PatchPortNotFound(OVNBGPAgentException):
    """Patch Port not found

    :param localnet: The localnet name
    """

    message = _("Patch port not found for localnet: %(localnet)s.")


class ExposeDeniedForAddressScope(OVNBGPAgentException):
    """Address Scope test failed

    :param addresses: The ip address used for checking address_scope
    :param address_scopes: The address scopes
    :param configured_scopes: The allowed address scopes in configuration
    """

    message = _("Exposing addresses %(addresses)s with address scopes "
                "%(address_scopes)s was denied, required scopes: "
                "%(configured_scopes)s")


class WireFailure(OVNBGPAgentException):
    """Wire port failed

    :param cidr: The cidr that failed to wire.
    :param message: The failure message
    """

    message = _("Failure with wiring for CIDR %(cidr)s: %(message)s")


class UnwireFailure(OVNBGPAgentException):
    """Unwire port failed

    :param cidr: The cidr that failed to wire.
    :param message: The failure message
    """

    message = _("Failure with removing wiring for CIDR %(cidr)s: %(message)s")


class IpAddressAlreadyExists(RuntimeError):
    message = _("IP address %(ip)s already configured on %(device)s.")

    def __init__(self, message=None, ip=None, device=None):
        message = message or self.message % {'ip': ip, 'device': device}
        super(IpAddressAlreadyExists, self).__init__(message)


class NetworkInterfaceNotFound(RuntimeError):
    message = _("Network interface %(device)s not found")

    def __init__(self, message=None, device=None):
        message = message or self.message % {'device': device}
        super(NetworkInterfaceNotFound, self).__init__(message)


class InterfaceAlreadyExists(RuntimeError):
    message = _("Interface %(device)s already exists.")

    def __init__(self, message=None, device=None):
        message = message or self.message % {'device': device}
        super(InterfaceAlreadyExists, self).__init__(message)


class InterfaceOperationNotSupported(RuntimeError):
    message = _("Operation not supported on interface %(device)s.")

    def __init__(self, message=None, device=None):
        message = message or self.message % {'device': device}
        super(InterfaceOperationNotSupported, self).__init__(message)


class InvalidArgument(RuntimeError):
    message = _("Invalid parameter/value used on interface %(device)s.")

    def __init__(self, message=None, device=None):
        message = message or self.message % {'device': device}
        super(InvalidArgument, self).__init__(message)


class MacAddressNotFound(OVNBGPAgentException):
    message = _("Addresses column is empty or not set for port %(lsp)s")


class IpAddressNotFound(OVNBGPAgentException):
    message = _("Addresses column has no IP addresses or is not set for port"
                " %(lsp)s")



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import common
from ovn_bgp_agent import exceptions


def has_ip_address_defined(address):
    return ' ' in address.strip()


def get_fip(lsp):
    return common.get_from_external_ids(lsp, key=constants.OVN_FIP_EXT_ID_KEY)


def has_additional_binding(row):
    # requested-chassis can be a comma separated list, so if there
    # is a comma in the string, there is an additional binding.
    return ',' in getattr(row, 'options', {}).get(
        constants.OVN_REQUESTED_CHASSIS, '')


def get_address_list(lsp):
    try:
        addrs = lsp.addresses[0].strip().split(' ')
        # Check the first element for an empty string
        if not addrs[0]:
            return []
    except (AttributeError, IndexError):
        return []

    return addrs


def get_mac_from_lsp(lsp):
    try:
        return get_address_list(lsp)[0]
    except IndexError:
        raise exceptions.MacAddressNotFound(lsp=lsp)


def get_ips_from_lsp(lsp):
    addresses = get_address_list(lsp)[1:]
    if not addresses:
        raise exceptions.IpAddressNotFound(lsp=lsp)

    return addresses


def make_lsp_dict(row):
    # TODO(jlibosva): Stop passing around dynamic maps
    return {
        'mac': get_mac_from_lsp(row),
        'cidrs': row.external_ids.get(constants.OVN_CIDRS_EXT_ID_KEY,
                                      "").split(),
        'type': row.type,
        'logical_switch': common.get_from_external_ids(
            row, constants.OVN_LS_NAME_EXT_ID_KEY),
    }


def make_lrp_dict(row):
    return {
        'mac': row.mac,
        'cidrs': row.networks,
        'type': constants.OVN_CR_LRP_PORT_TYPE,
        'logical_switch': common.get_from_external_ids(
            row, constants.OVN_LS_NAME_EXT_ID_KEY),
        'router': row.external_ids.get(constants.OVN_LR_NAME_EXT_ID_KEY),
    }
